from sets import Set
from itertools import chain, groupby

from mininet.topo import Topo
from mininet.node import Controller
from mininet.node import RemoteController
from mininet.node import OVSController
from mininet.log import info, error, debug, output, warn, setLogLevel
from mininet.examples.cluster import MininetCluster, SwitchBinPlacer, RoundRobinPlacer
from mininet.examples.clustercli import ClusterCLI as CLI
from mininet.net import Mininet

class CustomNet:

    """
    Parses SG topology and based on it builds mininet net.
    It adds all the needed nodes (hosts, switches) and links between them.

    Since link bandwidth limiting is done when the network is run, it is
    not done at this stage but later.
    """
    @classmethod
    def GenerateNet(cls, sg_topology, servers):
        """
        Creates MininetCluster instance based on SG topology class.
        This approach cannot use built-in placers and hopefully it 
        will be depricated.
        """
 
        cls.mininetTopology = sg_topology

        # Since we need to support multiple controllers, we cannot create
        # MininetTopology but we create empty cluster and add nodes to it.
        cls.net = MininetCluster(waitConnected=True, servers=servers)

        # Loop trough all the autonomous systems and their switches and add
        # them to the topology.
        for AS in sg_topology.AutonomousSystems:
            info(AS.name, "\n")
            for switch in AS.switches:
                mininetSwitch = cls.net.addSwitch(switch.name.encode('ascii'))
                info("Added switch: ", mininetSwitch, "to the network.\n")

        # Based on provided connectivity create links between connected
        # switches.
        for AS in sg_topology.AutonomousSystems:
            for link in AS.links:
                info( "Link: ", link[0].name, "<->", link[1].name, "\n")
                switch1 = [x for x in cls.net.switches if x.name == link[0].name]
                switch2 = [x for x in cls.net.switches if x.name == link[1].name]
                if switch1:
                    if switch2:
                        #switch1[0].linkTo(switch2[0])
                        cls.net.addLink(switch1[0], switch2[0], )
                        info("Added interal link: ", switch1[0].name, "<->", switch2[0].name, "\n")


        # Add link between autonomous systems edge switches.
        if sg_topology.links:
            for link in sg_topology.links:
                info( "Link: ", link[0].name, "<->", link[1].name, "\n")

                # Find edge switch in switches. We need it for the name.
                switch1 = [x for x in link[0].switches if x.edge == True]
                # Find edge switch in set of switched in net instance.
                switch1_1 = [x for x in cls.net.switches if x.name == switch1[0].name]

                # Find edge switch in switches. We need it for the name.
                switch2 = [x for x in link[1].switches if x.edge == True]
                # Find edge switch in set of switched in net instance.
                switch2_2 = [x for x in cls.net.switches if x.name == switch2[0].name]

                if switch1:
                    if switch2:
                        cls.net.addLink(switch1_1[0], switch2_2[0], )
                        info("Added AS link: ", switch1[0].name, "<->", switch2[0].name, "\n")


        switchToHostDict = dict()
        # Determine where to add hosts based on predefined flows and add them.
        if sg_topology.flows:
            switches = set([])
            for flow in sg_topology.flows:
                switches.add(flow.source[0])
                switches.add(flow.destination[0])

            for switch in switches:
                host = cls.net.addHost('h' + switch.name)
                s = [x for x in cls.net.switches if x.name == switch.name]
                cls.net.addLink(s[0], host,)
                switchToHostDict[switch.name] = host

        # Since we do not have single set of controllers (they are assigned to each switch)
        # we will create a set to ensure each one controller is only once:
        # 1. added to net,
        # 2, started.
        cls.controllers = set()
        # Each switch needs to be stated with assigned controller, so we remember controller
        # instance for convinience.
        cls.switch_to_controller = dict()
        for AS in sg_topology.AutonomousSystems:
            for switch in AS.switches:
                # Create set of controllers and add each one only once to net.
                if switch.controller not in cls.controllers:
                    debug("IP", switch.controller.ip, "port", switch.controller.port,"\n")
                    c = cls.net.addController(switch.controller.name, controller=RemoteController, ip=switch.controller.ip, port=switch.controller.port)
                    cls.controllers.add(switch.controller)
                cls.switch_to_controller[switch.name] = c

        cls.net.start()

        for c in cls.net.controllers:
            debug("Starting controller: ", c.name, "\n")
            c.start()

        for s in cls.net.switches:
            debug("Starting switch: ", s.name, "with controller: ", cls.switch_to_controller[s.name].name, "\n")
            s.start([cls.switch_to_controller[s.name]])

        return cls.net, switchToHostDict

    @classmethod
    def GenerateNet_2(cls, sg_topology, servers):
        """
        Creates MininetCluster instance based on SGMininetTopology class.
        This allows us to use built-in node placers.
        """
        cls.mininetTopology = sg_topology

        cls.topo = SGMininetTopology(sg_topology)

        # Since we need to support multiple controllers, we cannot create
        # MininetTopology but we create empty cluster and add nodes to it.
        cls.net = MininetCluster(waitConnected=True, servers=servers, topo=cls.topo, placement=RoundRobinPlacer, controller=None)

        switch_to_host_dict = dict()
        # Determine where to add hosts based on predefined flows and add them.
        if sg_topology.flows:
            switches = set([])
            for flow in sg_topology.flows:
                switches.add(flow.source[0])
                switches.add(flow.destination[0])

            for switch in switches:
                host = cls.net.getNodeByName('h' + switch.name)
                switch_to_host_dict[switch.name] = host

        cls.controllers = set()
        cont = set()
        # Each switch needs to be stated with assigned controller, so we remember controller
        # instance for convinience.
        cls.switch_to_controller = dict()
        for AS in sg_topology.AutonomousSystems:
            for switch in AS.switches:
                # Create set of controllers and add each one only once to net.
                if switch.controller not in cont:
                    debug("Adding distinct controller: ", switch.controller.name, "ip", switch.controller.ip, "port", switch.controller.port,"\n")
                    c = cls.net.addController(switch.controller.name, controller=RemoteController, ip=switch.controller.ip, port=switch.controller.port)
                    cls.controllers.add(c)
                    cont.add(switch.controller)
                cls.switch_to_controller[switch.name] = [x for x in cls.controllers if x.name == switch.controller.name][0]

        cls._start()

        return cls.net, switch_to_host_dict

    @classmethod
    def _start(cls):
        if not cls.net.built:
            cls.net.build()

        for c in cls.controllers:
            c.start()

        for s in cls.net.switches:
            debug("Starting switch: ", s.name, "with controller: ", cls.switch_to_controller[s.name].name, "\n")
            s.start([cls.switch_to_controller[s.name]])

        # Execute batch startup if existing. This should be part of the mininet API that can
        # be called after switchs startup (as called by net.startu()
        started = {}
        for swclass, switches in groupby(
                sorted( cls.net.switches, key=type ), type ):
            switches = tuple( switches )
            if hasattr( swclass, 'batchStartup' ):
                success = swclass.batchStartup( switches )
                started.update( { s: s for s in success } )

        cls.net.waitConnected()

    def GenerateNet_3(cls, sg_topology, servers):
        """
        Creates MininetCluster instance based on SGMininetTopology class.
        This allows us to use built-in node placers.
        """
        cls.mininetTopology = sg_topology

        cls.topo = SGMininetTopology(sg_topology)

        c = RemoteController(name="c1", ip="10.19.1.1", port=6000)

        # Since we need to support multiple controllers, we cannot create
        # MininetTopology but we create empty cluster and add nodes to it.
        cls.net = MininetCluster(waitConnected=True, servers=servers, topo=cls.topo, placement=RoundRobinPlacer, controller=c)

        switch_to_host_dict = dict()
        # Determine where to add hosts based on predefined flows and add them.
        if sg_topology.flows:
            switches = set([])
            for flow in sg_topology.flows:
                switches.add(flow.source[0])
                switches.add(flow.destination[0])

            for switch in switches:
                host = cls.net.getNodeByName('h' + switch.name)
                switch_to_host_dict[switch.name] = host

        cls.controllers = set()
        # Each switch needs to be stated with assigned controller, so we remember controller
        # instance for convinience.
        #cls.switch_to_controller = dict()
        #for AS in sg_topology.AutonomousSystems:
        #    for switch in AS.switches:
                # Create set of controllers and add each one only once to net.
        #        if switch.controller not in cls.controllers:
        #            debug("IP", switch.controller.ip, "port", switch.controller.port,"\n")
        #            c = cls.net.addController(switch.controller.name, controller=RemoteController, ip=switch.controller.ip, port=switch.controller.port)
        #            cls.controllers.add(c)
                #cls.switch_to_controller[switch.name] = RemoteController(name=switch.controller.name, ip=switch.controller.ip, port=switch.controller.port)
        #        cls.switch_to_controller[switch.name] = [x for x in cls.controllers if x.name == switch.controller.name][0]


        #c = RemoteController("c1", "10.19.1.1", 6000)

        print "CONTROLLERS"
        for c in cls.controllers:
            print "NAME: " + c.name

        #for c in cls.controllers:
        #    c.start()

        #for s in cls.net.switches:
        #    debug("Starting switch: ", s.name, "with controller: ", cls.switch_to_controller[s.name].name, "\n")
        #    s.start([cls.switch_to_controller[s.name]])
        #    s.start([c])

        cls.net.start()

        return cls.net, switch_to_host_dict



class SGMininetTopology( Topo ):
    """
    Parses SG topology and based on it builds mininet topology.
    It adds all the needed nodes (hosts, switches) and links between them.
    Since link bandwidth limiting is done when the network is run, it is
    not done at this stage but later.
    """
    switchToHostDict = None

    "Create custom topo."
    def __init__(self, sg_topology):
        """
        Constructor does all the actual work.
        """

        # Initialize topology
        Topo.__init__( self )

        # Loop trough all the autonomous systems and their switches and add 
        # them to the topology.
        for AS in sg_topology.AutonomousSystems:
            info(AS.name, "\n")
            for switch in AS.switches:
                mininetSwitch = self.addSwitch(switch.name.encode('ascii'))
                info("Added switch: ", mininetSwitch, "to the network.\n")

        # Based on provided connectivity create links between connected
        # switches.
        for AS in sg_topology.AutonomousSystems:
            for link in AS.links:
                info( "Link: ", link[0].name, "<->", link[1].name, "\n")
                switch1 = [x for x in AS.switches if x.name == link[0].name]
                switch2 = [x for x in AS.switches if x.name == link[1].name]
                if switch1:
                    if switch2:
                        self.addLink( switch1[0].name, switch2[0].name)
                        info("Added interal link: ", switch1[0].name, "<->", switch2[0].name, "\n")

        # Add link between autonomous systems edge switches
        if sg_topology.links:
            for link in sg_topology.links:
                info( "Link: ", link[0].name, "<->", link[1].name, "\n")
                switch1 = [x for x in link[0].switches if x.edge == True]
                switch2 = [x for x in link[1].switches if x.edge == True]
                if switch1:
                    if switch2:
                        self.addLink( switch1[0].name, switch2[0].name)
                        info("Added AS link: ", switch1[0].name, "<->", switch2[0].name, "\n")

        # Determine where to add hosts based on predefined flows and add them.
        self.switchToHostDict = dict()
        if sg_topology.flows:
            switches = set([])
            for flow in sg_topology.flows:
                switches.add(flow.source[0].name)
                switches.add(flow.destination[0].name)

            for switch in switches:
                host = self.addHost('h' + switch)
                self.addLink(host, switch)

                self.switchToHostDict[switch] = host

